'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var path = require('path');
var cosmiconfig = require('cosmiconfig');
var micromatch = require('micromatch');
var deepmerge = require('deepmerge');
var jsStringEscape = require('js-string-escape');
var api = require('@structured-types/api');
var types = require('./types.js');
var fs = require('fs');
var hostedGitInfo = require('hosted-git-info');
var parseRepositoryURL = require('@hutson/parse-repository-url');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

function _interopNamespace(e) {
  if (e && e.__esModule) return e;
  var n = Object.create(null);
  if (e) {
    Object.keys(e).forEach(function (k) {
      if (k !== 'default') {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () { return e[k]; }
        });
      }
    });
  }
  n["default"] = e;
  return Object.freeze(n);
}

var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var path__namespace = /*#__PURE__*/_interopNamespace(path);
var micromatch__default = /*#__PURE__*/_interopDefaultLegacy(micromatch);
var deepmerge__default = /*#__PURE__*/_interopDefaultLegacy(deepmerge);
var jsStringEscape__default = /*#__PURE__*/_interopDefaultLegacy(jsStringEscape);
var fs__namespace = /*#__PURE__*/_interopNamespace(fs);
var hostedGitInfo__default = /*#__PURE__*/_interopDefaultLegacy(hostedGitInfo);
var parseRepositoryURL__default = /*#__PURE__*/_interopDefaultLegacy(parseRepositoryURL);

const traverseFolder = (filePath, levels = 10, fileName = 'package.json') => {
    if (!fs__namespace.existsSync(filePath)) {
        return null;
    }
    const files = fs__namespace.readdirSync(filePath);
    if (levels === 0) {
        return null;
    }
    const pckg = files.find((file) => file === fileName);
    if (pckg) {
        return path__default["default"].resolve(filePath, pckg);
    }
    return traverseFolder(path__default["default"].resolve(filePath, '..'), levels - 1, fileName);
};
/**
 * Retrieves the repo path from the project's package.json file.
 * @param filePath file path to start the search for a package.json
 */
const getRepoPath = (filePath) => {
    const packageFileName = traverseFolder(path__default["default"].dirname(filePath));
    const result = {};
    if (packageFileName) {
        const packageJSON = JSON.parse(fs__namespace.readFileSync(packageFileName, 'utf8'));
        result.packageName = packageJSON.name;
        result.relativePath = path__default["default"].relative(path__default["default"].dirname(packageFileName), filePath);
        const repositoryURL = typeof packageJSON.repository === 'string'
            ? packageJSON.repository
            : packageJSON.repository && packageJSON.repository.url;
        if (repositoryURL) {
            const templates = hostedGitInfo__default["default"].fromUrl(repositoryURL) ||
                parseRepositoryURL__default["default"](repositoryURL);
            result.relativePath = `${packageJSON.repository && packageJSON.repository.directory
                ? `${packageJSON.repository.directory}/`
                : ''}${path__default["default"].relative(path__default["default"].dirname(packageFileName), filePath)}`;
            const fillTemplate = (template) => {
                return template
                    .replace('auth@', templates.auth || '')
                    .replace('{domain}', templates.domain || '')
                    .replace('{user}', templates.user || '')
                    .replace('{project}', templates.project || '')
                    .replace('{#fragment}', '')
                    .replace('{path}', result.relativePath)
                    .replace('{treepath}', templates.treepath || '')
                    .replace('{/tree/committish}', `/${templates.treepath}/${templates.committish || 'master'}`)
                    .replace('{committish}', templates.committish || 'master');
            };
            result.repo = fillTemplate(templates.browsefiletemplate);
        }
    }
    return result;
};

const textNode = (value) => ({ kind: types.NodeKind.Text, value });

const tableCellNode = (value) => ({
    kind: types.NodeKind.TableCell,
    children: typeof value === 'string' ? [textNode(value)] : value,
});

const inlineCodeNode = (value) => value
    ? {
        kind: types.NodeKind.InlineCode,
        value,
    }
    : {
        kind: types.NodeKind.Text,
        value,
    };

const createPropsRow = ({ name, type, default: defaultValue, parents, description, prop }, columnOptions, visColumns) => {
    const getCell = (column) => {
        switch (column) {
            case 'name':
                return name;
            case 'type':
                return type;
            case 'default':
                return [
                    inlineCodeNode(typeof defaultValue !== 'undefined' ? defaultValue.toString() : ''),
                ];
            case 'description':
                if (typeof description === 'string') {
                    const parts = description.split('`');
                    if (parts.length > 1) {
                        return parts.reduce((acc, text, idx) => {
                            if (idx % 2 === 0) {
                                return [...acc, textNode(text)];
                            }
                            else {
                                return [
                                    ...acc,
                                    textNode(' '),
                                    inlineCodeNode(text),
                                    textNode(' '),
                                ];
                            }
                        }, []);
                    }
                    else {
                        return [textNode(description)];
                    }
                }
                else {
                    return [textNode('')];
                }
            case 'parents':
                return parents;
            default:
                return undefined;
        }
    };
    const children = [];
    Object.keys(columnOptions).forEach((key) => {
        const name = key;
        const column = columnOptions[name];
        if (column && visColumns[name]) {
            let cell = undefined;
            if (typeof column.render === 'function') {
                cell = column.render(name, prop);
            }
            if (!cell) {
                const content = getCell(name);
                cell = typeof content === 'string' ? textNode(content) : content;
            }
            if (cell) {
                if (Array.isArray(cell)) {
                    children.push(tableCellNode(cell));
                }
                else {
                    children.push(tableCellNode([cell]));
                }
            }
        }
    });
    return {
        kind: types.NodeKind.TableRow,
        children,
    };
};
const createPropsTable = (children, columnOptions, prop) => {
    var _a;
    const propsTable = [];
    let table = undefined;
    const visibleColumns = {
        name: false,
        default: false,
        description: false,
        parents: false,
        type: true,
    };
    const getTitle = (config, prop) => {
        let value = '';
        const { title } = config;
        if (title) {
            value = (typeof title === 'function' ? title(prop) : title) || '';
        }
        return tableCellNode([textNode(value)]);
    };
    if (children) {
        const columns = [];
        Object.keys(columnOptions).forEach((key) => {
            const name = key;
            const column = columnOptions[name];
            if (column && !column.hidden) {
                visibleColumns[name] = children.some((item) => item[name] !== undefined);
                if (visibleColumns[name]) {
                    columns.push(getTitle(column, prop));
                }
            }
        });
        table = {
            kind: types.NodeKind.Table,
            children: [
                {
                    kind: types.NodeKind.TableRow,
                    children: columns,
                },
            ],
        };
        propsTable.push(table);
        (_a = table.children) === null || _a === void 0 ? void 0 : _a.push(...children.map((child) => createPropsRow(child, columnOptions, visibleColumns)));
    }
    return { propsTable: propsTable, table, visibleColumns };
};

const paragraphNode = (value) => ({
    kind: types.NodeKind.Paragraph,
    children: typeof value === 'string' ? [textNode(value)] : value,
});

const emphasisNode = (value) => ({
    kind: types.NodeKind.Emphasis,
    children: typeof value === 'string' ? [textNode(value)] : value,
});

const codeNode = (value) => ({
    kind: types.NodeKind.Code,
    value,
});

const linkNode = (children, url, loc) => {
    const prop = {
        kind: types.NodeKind.Link,
        children,
    };
    if (url) {
        prop.url = url;
    }
    if (loc) {
        prop.loc = loc;
    }
    return prop;
};

const headingNode = (value, depth) => ({
    kind: types.NodeKind.Heading,
    depth,
    children: typeof value === 'string' ? [textNode(value)] : value,
});

const boldNode = (value) => ({
    kind: types.NodeKind.Bold,
    children: typeof value === 'string' ? [textNode(value)] : value,
});

const defaultSections = {
    title: {},
    type: {},
    description: {},
    location: {},
    extends: { title: 'extends' },
    props: { title: 'properties' },
    examples: {
        title: (prop) => prop.examples && prop.examples.length > 1 ? 'examples' : 'example',
    },
};

const defaultColumns = {
    name: { title: 'Name' },
    type: { title: 'Type' },
    parents: { title: 'Parent' },
    default: { title: 'Default' },
    description: { title: 'Description' },
};

/* eslint-disable prefer-spread */
class PropsToDocumentation {
    constructor() {
        this.sections = defaultSections;
        this.columns = defaultColumns;
        this.topLevelProps = {};
        this.helpers = {};
        this.options = {};
        this.repoNames = {};
        this.getPropLink = (key) => {
            const nameParts = key.split('.');
            return this.topLevelProps[nameParts[nameParts.length - 1]];
        };
    }
    extractPropTable(prop, rows) {
        let parentProp = undefined;
        const addParentProp = (prop) => {
            var _a, _b;
            if (prop.parent) {
                if (!parentProp) {
                    parentProp = {
                        name: '...props',
                        kind: api.PropKind.Enum,
                        properties: [
                            {
                                kind: api.PropKind.Type,
                                type: prop.parent.name,
                                loc: prop.parent.loc,
                            },
                        ],
                        optional: true,
                    };
                }
                else {
                    if (!((_a = parentProp.properties) === null || _a === void 0 ? void 0 : _a.find((p) => { var _a; return p.type === ((_a = prop.parent) === null || _a === void 0 ? void 0 : _a.name); }))) {
                        (_b = parentProp.properties) === null || _b === void 0 ? void 0 : _b.push({
                            kind: api.PropKind.Type,
                            type: prop.parent.name,
                            loc: prop.parent.loc,
                        });
                    }
                }
            }
        };
        const consolidatedProps = rows.filter((prop) => {
            var _a, _b;
            if (prop.parent) {
                if (this.options.skipInherited ||
                    ((_a = this.options.collapsed) === null || _a === void 0 ? void 0 : _a.includes(prop.parent.name))) {
                    addParentProp(prop);
                    return false;
                }
                if (this.options.collapsed) {
                    for (const collapsedProp of this.options.collapsed) {
                        const helperParent = this.helpers[collapsedProp];
                        if (helperParent && api.hasProperties(helperParent)) {
                            const helpProp = (_b = helperParent.properties) === null || _b === void 0 ? void 0 : _b.find((p) => { var _a, _b; return p.name === prop.name && ((_a = p.parent) === null || _a === void 0 ? void 0 : _a.name) === ((_b = prop.parent) === null || _b === void 0 ? void 0 : _b.name); });
                            if (helpProp) {
                                addParentProp({
                                    ...prop,
                                    parent: { name: collapsedProp, loc: helperParent.loc },
                                });
                                return false;
                            }
                        }
                    }
                }
            }
            return true;
        });
        const allProps = parentProp
            ? [...consolidatedProps, parentProp]
            : consolidatedProps;
        const items = allProps.map((prop) => {
            let name = prop.name
                ? prop.kind === api.PropKind.Rest
                    ? `...${prop.name}`
                    : prop.name
                : '';
            name = `${name}${!name || prop.optional ? '' : '*'}`;
            return this.configurePropItem({
                name: prop.loc
                    ? this.propLink({ name, loc: prop.loc })
                    : [inlineCodeNode(name)],
                parents: prop.parent ? this.propLink(prop.parent) : undefined,
                type: this.extractPropType(prop, { extractProperties: true }),
                description: prop.description,
                default: prop.value,
                prop,
            });
        });
        return createPropsTable(items, this.columns, prop);
    }
    extractFunctionDeclaration(prop) {
        const result = [textNode('(')];
        if (prop.parameters) {
            for (let i = 0; i < prop.parameters.length; i += 1) {
                const p = prop.parameters[i];
                if (i > 0) {
                    result.push(textNode(', '));
                }
                if (p.name) {
                    result.push(inlineCodeNode(p.name));
                    if (!p.optional) {
                        result.push(textNode('*'));
                    }
                    result.push(textNode(': '));
                }
                result.push(...this.extractPropType(p));
                if (!p.name && !p.optional) {
                    result.push(textNode('*'));
                }
            }
        }
        result.push(textNode(')'));
        result.push(textNode(' => '));
        if (prop.returns) {
            result.push(...this.extractPropType(prop.returns));
        }
        else {
            result.push(textNode('void'));
        }
        return result;
    }
    configurePropItem(item) {
        const enabledColumn = (name) => {
            var _a;
            return this.columns[name] && !((_a = this.columns[name]) === null || _a === void 0 ? void 0 : _a.hidden);
        };
        return {
            name: enabledColumn('name') ? item.name : undefined,
            parents: enabledColumn('parents') ? item.parents : undefined,
            type: enabledColumn('type') ? item.type : undefined,
            default: enabledColumn('default') ? item.default : undefined,
            description: enabledColumn('description') ? item.description : undefined,
            prop: item.prop,
        };
    }
    extractFunction(prop, _extractTable = true) {
        if (prop.parameters) {
            const { propsTable, table, visibleColumns } = this.extractPropTable(prop, prop.parameters);
            if (table &&
                table.children &&
                prop.returns &&
                prop.returns.kind !== api.PropKind.Void) {
                table.children.push(createPropsRow(this.configurePropItem({
                    name: [inlineCodeNode('returns')],
                    parents: prop.returns.parent
                        ? this.propLink(prop.returns.parent)
                        : undefined,
                    type: this.extractPropType(prop.returns),
                    description: prop.returns.description,
                    default: undefined,
                    prop,
                }), this.columns, visibleColumns));
            }
            return propsTable;
        }
        return undefined;
    }
    extractInterface(prop) {
        const result = [];
        if (prop.name) {
            if (api.isUnionProp(prop)) {
                result.push(...this.extractPropType(prop));
            }
            else if (api.hasProperties(prop) && prop.properties) {
                const { propsTable } = this.extractPropTable(prop, prop.properties);
                result.push(...propsTable);
            }
        }
        return result;
    }
    propLink(prop) {
        var _a;
        const typeText = [inlineCodeNode(prop.name)];
        if (typeof prop.name === 'string') {
            const link = this.getPropLink(prop.name);
            return [
                linkNode(typeText, link ? `#${(_a = link.name) === null || _a === void 0 ? void 0 : _a.toLowerCase()}` : undefined, prop.loc),
            ];
        }
        return typeText;
    }
    inlineType(prop) {
        let typeNode = undefined;
        if (typeof prop.type === 'string') {
            if (this.getPropLink(prop.type)) {
                typeNode = this.propLink({ name: prop.type, loc: prop.loc });
            }
            else {
                typeNode = this.extractPropType(prop);
            }
        }
        else if (prop.kind) {
            typeNode = this.extractPropType(prop);
        }
        if (prop.name && prop.name !== prop.type) {
            return [
                inlineCodeNode(`${prop.name}`),
                textNode(`${typeNode ? ': ' : ''}`),
                ...(typeNode || []),
            ];
        }
        return typeNode || [];
    }
    typeNode(prop, showValue = true) {
        if (typeof prop.type === 'string') {
            if (typeof prop.parent === 'string') {
                return [...this.propLink(prop.parent), textNode(`.${prop.type}`)];
            }
            return this.propLink({ name: prop.type, loc: prop.loc });
        }
        if (showValue && prop.value !== undefined) {
            const value = api.isStringProp(prop)
                ? `"${jsStringEscape__default["default"](prop.value)}"`
                : prop.value.toString();
            return [textNode(value)];
        }
        if (prop.kind) {
            const typeNode = [
                inlineCodeNode(`${api.PropKind[prop.kind].toLowerCase()}`),
            ];
            if (typeof prop.parent === 'string' && this.getPropLink(prop.parent)) {
                const link = this.propLink(prop.parent);
                if (link.length) {
                    typeNode.push(textNode(` (`));
                    typeNode.push(...link);
                    typeNode.push(textNode(`)`));
                }
            }
            return typeNode;
        }
        if (prop.name) {
            return [textNode(prop.name)];
        }
        return [];
    }
    extractPropType(prop, options) {
        var _a;
        if (prop.parent) {
            const parent = this.getPropLink(prop.parent.name);
            if (parent && api.isClassLikeProp(parent)) {
                const p = (_a = parent.properties) === null || _a === void 0 ? void 0 : _a.find((p) => p.name === prop.name);
                if (p) {
                    return this.extractType(p, options);
                }
            }
        }
        return this.extractType(prop, options);
    }
    getPropProperties(prop) {
        const { maxProps = 30 } = this.options;
        if (prop.properties) {
            return {
                properties: maxProps ? prop.properties.slice(0, maxProps) : [],
                more: maxProps && prop.properties.length > maxProps
                    ? textNode(` ,...${prop.properties.length - maxProps} more`)
                    : undefined,
            };
        }
        return {
            properties: [],
            more: undefined,
        };
    }
    extractType(prop, options) {
        var _a, _b, _c;
        const unionSeparator = ' | ';
        const enumSeparator = ' | ';
        const { showValue = false } = options || {};
        if (typeof prop.type === 'string' &&
            ((_a = this.options.collapsed) === null || _a === void 0 ? void 0 : _a.includes(prop.type))) {
            return this.typeNode(prop, showValue);
        }
        else if (api.isUnionProp(prop) || api.isEnumProp(prop)) {
            const separator = api.isUnionProp(prop) ? unionSeparator : enumSeparator;
            const { properties, more } = this.getPropProperties(prop);
            const nodes = properties.reduce((acc, t, idx) => {
                const r = [...acc, ...this.extractPropType(t, { showValue: true })];
                if (idx < properties.length - 1) {
                    r.push(textNode(separator));
                }
                return r;
            }, []);
            if (more) {
                nodes.push(more);
            }
            return nodes;
        }
        else if (api.isClassLikeProp(prop)) {
            const propName = typeof prop.type === 'string' ? prop.type : prop.name;
            const result = [];
            const { properties, more } = this.getPropProperties(prop);
            if (typeof propName === 'string' && this.getPropLink(propName)) {
                result.push(...this.propLink({ name: propName, loc: prop.loc }));
            }
            else if (properties.length) {
                const typeArguments = properties.reduce((acc, p, idx) => {
                    const result = [...acc, ...this.inlineType(p)];
                    if (properties && idx < properties.length - 1) {
                        result.push(textNode(', '));
                    }
                    return result;
                }, []);
                if (more) {
                    typeArguments.push(more);
                }
                if (typeArguments.length) {
                    result.push(textNode('{ '));
                    result.push(...typeArguments);
                    result.push(textNode(' }'));
                }
            }
            else if ((_b = prop.generics) === null || _b === void 0 ? void 0 : _b.length) {
                const genericArguments = (_c = prop.generics) === null || _c === void 0 ? void 0 : _c.reduce((acc, p, idx) => {
                    const result = [...acc, ...this.inlineType(p)];
                    if (prop.generics && idx < prop.generics.length - 1) {
                        result.push(textNode(', '));
                    }
                    return result;
                }, []);
                result.push(...this.typeNode(prop));
                if (genericArguments === null || genericArguments === void 0 ? void 0 : genericArguments.length) {
                    result.push(...[textNode('<'), ...genericArguments, textNode('>')]);
                }
            }
            else if (propName) {
                result.push(textNode(propName));
            }
            return result;
        }
        else if (api.isArrayProp(prop)) {
            const { properties, more } = this.getPropProperties(prop);
            const elements = properties.reduce((acc, p, idx) => {
                const result = this.extractPropType(p);
                if (idx < properties.length - 1) {
                    result.push(textNode(', '));
                }
                return [...acc, ...result];
            }, []);
            if (more) {
                elements.push(more);
            }
            const multiProps = elements.length &&
                ((elements[0].children && elements[0].children.length > 1) ||
                    (elements.length > 2 &&
                        types.isNodeWithValue(elements[1]) &&
                        (elements[1].value === unionSeparator ||
                            elements[1].value === enumSeparator)));
            if (multiProps) {
                elements.splice(0, 0, textNode('('));
                elements.push(textNode(')'));
            }
            elements.push(textNode('[]'));
            return elements;
        }
        else if (api.isTupleProp(prop)) {
            const { properties, more } = this.getPropProperties(prop);
            const nodes = [
                textNode('['),
                ...properties.reduce((acc, p, idx) => {
                    const result = this.extractPropType(p);
                    if (idx < properties.length - 1) {
                        result.push(textNode(', '));
                    }
                    return [...acc, ...result];
                }, []),
            ];
            if (more) {
                nodes.push(more);
            }
            nodes.push(textNode(']'));
            return nodes;
        }
        else if (api.isIndexProp(prop)) {
            const results = [
                textNode('['),
                ...this.extractPropType(prop.index),
                textNode(']'),
            ];
            if (prop.prop) {
                results.push(textNode(': '));
                results.push(...this.extractPropType(prop.prop));
            }
            return results;
            //return this.extractFunctionDeclaration(prop);
        }
        else if (api.isFunctionProp(prop)) {
            return this.extractFunctionDeclaration(prop);
        }
        return this.typeNode(prop, showValue);
    }
    getSourceLocation(prop) {
        const { filePath, loc } = prop.loc || {};
        if (filePath) {
            if (!this.repoNames[filePath]) {
                this.repoNames[filePath] = getRepoPath(path__default["default"].resolve(filePath));
            }
            if (this.repoNames[filePath]) {
                const { repo, relativePath, packageName } = this.repoNames[filePath];
                const fileLocation = repo || filePath;
                if (fileLocation) {
                    const { line } = (loc === null || loc === void 0 ? void 0 : loc.start) || {};
                    const sourceLocation = filePath.includes('node_modules')
                        ? fileLocation
                        : `${fileLocation}${line ? `#L${line}` : ''}`;
                    return [
                        paragraphNode([
                            emphasisNode([
                                textNode('defined in '),
                                linkNode([
                                    textNode(packageName
                                        ? `${packageName}/${relativePath}`
                                        : sourceLocation),
                                ], sourceLocation, prop.loc),
                            ]),
                        ]),
                    ];
                }
            }
        }
        return [];
    }
    getType(prop) {
        if (prop.kind) {
            return [
                boldNode([
                    inlineCodeNode(`${prop.extension ? `${prop.extension} ` : ''}${api.PropKind[prop.kind].toLowerCase()}`),
                ]),
            ];
        }
        else if (typeof prop.type === 'string') {
            return [
                headingNode(this.propLink({ name: prop.type, loc: prop.loc }), 3),
            ];
        }
        return undefined;
    }
    getLocation(prop) {
        const loc = this.getSourceLocation(prop);
        if (loc.length) {
            return loc;
        }
        return undefined;
    }
    getPropsTable(prop) {
        if (api.isFunctionProp(prop)) {
            return this.extractFunction(prop);
        }
        else if (api.hasProperties(prop)) {
            return this.extractInterface(prop);
        }
        return undefined;
    }
    getExamples(prop) {
        if (prop.examples) {
            const result = [];
            const codeExamples = prop.examples.filter((e) => e.content);
            codeExamples.forEach((example) => {
                if (example.content) {
                    result.push(paragraphNode([codeNode(example.content)]));
                }
            });
            return result;
        }
        return undefined;
    }
    getExtends(prop) {
        var _a;
        if (api.isClassLikeProp(prop) && ((_a = prop.extends) === null || _a === void 0 ? void 0 : _a.length)) {
            const extendsList = prop.extends.reduce((acc, key, idx) => {
                const p = this.getPropLink(key);
                let result;
                if (p) {
                    result = this.extractPropType(p);
                }
                else {
                    result = [textNode(key)];
                }
                if (prop.extends && idx < prop.extends.length - 1) {
                    result.push(textNode(', '));
                }
                return [...acc, ...result];
            }, []);
            return extendsList;
        }
        return undefined;
    }
    getSection(prop, name) {
        switch (name) {
            case 'title':
                return prop.name ? [headingNode(prop.name, 2)] : undefined;
            case 'description':
                return prop.description
                    ? [
                        paragraphNode(prop.description.split('\n').map((d) => ({
                            kind: types.NodeKind.Text,
                            value: d,
                        }))),
                    ]
                    : undefined;
            case 'type':
                return this.getType(prop);
            case 'location':
                return this.getLocation(prop);
            case 'props':
                return this.getPropsTable(prop);
            case 'extends':
                return this.getExtends(prop);
            case 'examples':
                return this.getExamples(prop);
            default:
                return undefined;
        }
    }
    generateSection(prop, name) {
        const config = this.sections[name];
        if (config) {
            const { title, render } = config;
            const children = typeof render === 'function'
                ? render(prop)
                : this.getSection(prop, name);
            if (children) {
                const result = [];
                if (title) {
                    const value = typeof title === 'function' ? title(prop) : title;
                    if (value) {
                        result.push(paragraphNode([boldNode([textNode(value)])]));
                    }
                }
                result.push(...children);
                return result;
            }
        }
        return undefined;
    }
    generateNodes(prop) {
        const result = [];
        Object.keys(this.sections).forEach((name) => {
            const config = this.sections[name];
            if (!(config === null || config === void 0 ? void 0 : config.hidden)) {
                const section = this.generateSection(prop, name);
                if (section) {
                    result.push(...section);
                }
            }
        });
        return result;
    }
    extract(props, options = {}) {
        const result = [];
        const { extensions, visible, sections = {}, columns = {} } = options;
        this.options = options;
        if (sections) {
            this.sections = Array.isArray(sections)
                ? sections.reduce((acc, name) => {
                    return { ...acc, [name]: defaultSections[name] };
                }, {})
                : deepmerge__default["default"](defaultSections, sections);
        }
        if (columns) {
            this.columns = Array.isArray(columns)
                ? columns.reduce((acc, name) => {
                    return { ...acc, [name]: defaultColumns[name] };
                }, {})
                : deepmerge__default["default"](defaultColumns, columns);
        }
        const filterProps = (prop) => {
            if (Array.isArray(extensions)) {
                if (typeof prop.extension === 'string' &&
                    extensions.includes(prop.extension)) {
                    return prop;
                }
                return undefined;
            }
            if (Array.isArray(visible)) {
                if (typeof prop.name === 'string' && visible.includes(prop.name)) {
                    return prop;
                }
                return undefined;
            }
            return prop;
        };
        Object.keys(props).forEach((key) => {
            if (key !== '__helpers' && key !== '__diagnostics') {
                const prop = filterProps(props[key]);
                if (prop) {
                    this.topLevelProps[key] = prop;
                }
            }
        });
        this.helpers = props.__helpers || {};
        Object.keys(this.helpers).forEach((key) => {
            const prop = filterProps(this.helpers[key]);
            if (prop) {
                this.topLevelProps[key] = prop;
            }
        });
        Object.values(this.topLevelProps).forEach((prop) => {
            const nodes = this.generateNodes(prop);
            result.push(...nodes);
        });
        return result;
    }
}

/**
 * Creates a list of api documentation nodes
 * @param props properties parsed from `structured-types/api`
 * @param options page generation options
 * @returns a list of documentation nodes
 */
const propsToDocumentation = (props, options) => {
    const extractor = new PropsToDocumentation();
    return extractor.extract(props, options);
};
/**
 * deep merge of documentation configurations, by replacing array members (ie sections, columns)
 */
const mergeConfig = (dest, src) => deepmerge__default["default"](dest, src, {
    arrayMerge: (dest, src) => src,
});
/**
 * Read the api-docs configuration file
 * @param fileName the file that is being analyzed, will be used the starting folder to search for configuration files.
 * @param configFileName pass directly the configuration file name
 * @param elementId an optional element id. If not, the elements will be micromatch'ed based on relative path from configuration
 * @returns page generation options from the config file
 */
const apiDocsConfig = (fileName, configFileName, elementId) => {
    const configExplorer = cosmiconfig.cosmiconfigSync('api-docs');
    let results;
    if (configFileName) {
        results = configExplorer.load(configFileName);
    }
    else {
        const searchPath = path__namespace.dirname(fileName);
        results = configExplorer.search(searchPath);
    }
    if (results) {
        const { filepath, config } = results;
        if (filepath && config) {
            const { elements, ...rest } = config;
            results.config = rest;
            if (elements) {
                const id = elementId || path__namespace.relative(path__namespace.dirname(filepath), fileName);
                const onlyName = path__namespace.basename(fileName);
                Object.keys(elements).forEach((key) => {
                    if (onlyName === key || micromatch__default["default"].isMatch(id, key)) {
                        results.config = mergeConfig(results.config, elements[key]);
                    }
                });
            }
        }
    }
    return results;
};

Object.defineProperty(exports, 'NodeKind', {
  enumerable: true,
  get: function () { return types.NodeKind; }
});
exports.isBoldNode = types.isBoldNode;
exports.isCodeNode = types.isCodeNode;
exports.isEmphasisNode = types.isEmphasisNode;
exports.isHeadingNode = types.isHeadingNode;
exports.isInlineCodeNode = types.isInlineCodeNode;
exports.isLinkNode = types.isLinkNode;
exports.isNodeWithChildren = types.isNodeWithChildren;
exports.isNodeWithValue = types.isNodeWithValue;
exports.isParagraphNode = types.isParagraphNode;
exports.isTableCellNode = types.isTableCellNode;
exports.isTableNode = types.isTableNode;
exports.isTableRowNode = types.isTableRowNode;
exports.isTextNode = types.isTextNode;
exports.apiDocsConfig = apiDocsConfig;
exports.getRepoPath = getRepoPath;
exports.mergeConfig = mergeConfig;
exports.propsToDocumentation = propsToDocumentation;
